/* 
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_pthread.c,v 1.26 2004/10/19 22:10:58 phargrov Exp $
 */

#include <string.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <errno.h>
#include "cr_private.h"

// Holds thread-specific data key for cri_info
pthread_key_t cri_info_key;

// atfork callback to reset the state
static void
child_reset(void)
{
    cri_rb_init(&cri_cs_lock);
    cri_atomic_write(&cri_live_count, 0);
    cri_info_free(cri_info_location());
}

// Interpose on fork() to ensure the pthreads version is called
// if the caller linked us, but not the pthreads library.
pid_t cri_fork(void)
{
  return __fork();
}
weak_alias(cri_fork,fork);

//
// Initialize pthread-dependent parts
//
void
cri_pthread_init(void)
{
    int rc;

    // See bug #460
    {
	void *handle = dlopen(NULL /* the executable */, RTLD_LAZY);
	if (dlsym(handle, "__fork") == dlsym(handle, "__libc_fork")) {
	    CRI_ABORT("Unable to ensure pthread_atfork() will run.\n"
		      "If you dlopen(libcr.so) or use older libraries/loaders, "
		      "then you may need to link libpthread explicitly (or use LD_PRELOAD).");
	}
	dlclose(handle);
    }

    // Install a pthread_atfork callback to cleanup state in children
    rc = pthread_atfork(NULL, NULL, &child_reset);
    if (rc != 0) {
	CRI_ABORT("pthread_atfork() returned %d", rc);
    }

    // Setup the thread-specific data for cri_info
    rc = pthread_key_create(&cri_info_key, &cri_info_free);
    if (rc != 0) {
	CRI_ABORT("pthread_key_create() returned %d", rc);
    }
}
